/*
Murasaki - multiple genome global alignment program
Copyright (C) 2006-2007 Keio University
(Kris Popendorf) <comp@bio.keio.ac.jp> (2006)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

////////
// sequence defs
// Kris Popendorf
///////

#ifndef SEQUENCE__H
#define SEQUENCE__H

#include "cmultiset.h"

#include <iostream>
#include <vector>
#include <set>
#include <map>
#include <list>
#include <limits.h>
#include <functional>
#include <assert.h>
#include "itree.h"
#include "seqread.h"

#ifdef MURASAKI_SQLITE3
#include <sqlite3.h>
#endif

#ifdef MURASAKI_MYSQL
#include <mysql.h>
#endif

#ifdef USE_SHM_SYSV
#include <sys/types.h>
#include <sys/ipc.h>
#endif

using namespace std;
using namespace Itree;

//in my world:
// C = 00
// A = 01
// T = 10
// G = 11
// isnt that cute? ~G==C and so on

#ifdef __linux__
#define WORDSIZE __WORDSIZE
#else
#if (defined(__FreeBSD__) && defined(__i386__)) || (defined (__APPLE__) && (defined (__ppc__) || defined (__i386__)))
#define WORDSIZE 32
#endif

#if (defined (__APPLE__) && defined (__ppc64__)) || (defined(__FreeBSD__) && defined(__x86_64__))
#define WORDSIZE 64
#endif
#endif

#define MODWORDSIZE(a)  ((a) & (WORDSIZE-1))

typedef unsigned long word;
#ifdef LARGESEQ_SUPPORT
typedef long SeqPos;
#else
typedef int SeqPos;
#endif

extern int _wordsize,hash_max; 
extern int max_hash_bits;//hash_max limits memory usage
extern word hash_size; //actually used size, might be less than max
extern int mismatch_penalty,match_score,max_credits;
extern bool gappingEnabled;
typedef word HashKey;
typedef map<word,SeqPos> HashCount;

enum BASE_TYPE {BASE_C,BASE_A,BASE_T,BASE_G,BASES};
extern double globalBaseFreq[4];
extern word globalBaseCount[4],globalCounted;
enum SharedMemType {SHM_NONE=0,SHM_MMAP_RO, SHM_MMAP_RW, SHM_SYSV, SHM_END};

void finishGlobalFrequencies();

//forward declarations...
class HashMethod;
class Window;
class Anchor;
class Location;
class AnchorSet;
class UsedMap;
class Sequence;
class BitSequence;

void set_seqtable(vector<Sequence*> seqs);

void initHashParams(int bits,BitSequence *pat);
void initConstants();

inline word revComp8bit(word t);
inline word revCompWord(word t);
inline int highestBit(word t);
inline word lowN(int i);
inline word highN(int i);
inline word first2(word w){return w>>(WORDSIZE-2);}

int popCount(word w);
string wordToString(word w,int bits=WORDSIZE);
string wordToPattern(word w,int bits=WORDSIZE);

string hashCacheName(string name, BitSequence *pat);

//bits for sequence index table
typedef unsigned char SeqIdx;
extern map<Sequence*,SeqIdx> seq2idx;

typedef itree<SeqPos,AnchorSet*> usedItree;
typedef usedItree::iterator MapIte;
typedef Interval<SeqPos> UsedInt;

extern int next_sequence_;
extern int base_patt_len;
extern word activeHash;
extern word totalSequenceMemory;
extern word sys_pagesize;

//for managing UsedInts
UsedInt UsedInt_grown(const UsedInt &a,const SeqPos amt);
UsedInt UsedInt_inverted(const UsedInt &a,const Sequence& s);
UsedInt& UsedInt_invert(UsedInt &a,const Sequence& s);
UsedInt UsedInt_coalesced(const UsedInt &a,const UsedInt &b);
void UsedInt_coalesce(UsedInt &x,const UsedInt &a);
bool UsedInt_contains(const UsedInt &outer,const UsedInt &inner);
string asAnchor(const UsedInt& a,const Sequence *s);
bool UsedInt_sanityCheck(const UsedInt &a);

//inline function defs actually have to be in here if they're used in other .cc's...
const inline SeqIdx lookup_seqtable(Sequence* s) {
  assert(seq2idx.find(s)!=seq2idx.end()); //must be in there already
  return seq2idx[s];
}

inline SeqPos bitSeqCoords(SeqPos seqCoords){
  return abs(seqCoords)-1;
}

inline word lowN(int i){
  return ~(((word)-1)<<i); //lowest n bits are 1
}

inline word highN(int i){
  return ~((~(word)0)>>i);
}

class BitSequence { //for packing dna strings into 2 bit chunks
 public:
  //accessor
  inline const word* words() const {return _words;}
  inline word wordCount() const {return word_count;}
  inline word reverseOtf(SeqPos pos) const;
  inline SeqPos length(){return _length;}
  int wordsize();
  string asString() const;
  string asPattern() const;
  inline int readCode(SeqPos pos) const;
  inline word readWord(SeqPos pos) const;
  inline SeqPos readBitCount(){return bit_count;}
  inline int hashLength(){return _hashLength;}
  int compHashLength();
  pair<SeqPos,SeqPos>* localRegion(SeqPos pos); //find region local to pos
  SeqPos spaceRight(SeqPos pos); //matchable space right of pos
  SeqPos spaceLeft(SeqPos pos); //matchable space left of pos
  string genBitFilename();

  SeqPos cmpRight(SeqPos pos,BitSequence& target,SeqPos targetPos);
  SeqPos cmpLeft(SeqPos pos,BitSequence& target,SeqPos targetPos);

  //manipulate others
  void maskString(string &str) const;

  //operator
  bool equal(SeqPos pos,BitSequence& target,SeqPos targetPos,BitSequence& patt);

  //mutator
  void invert(); //flips to comp sequence
  
  //bitsequence constructors
  BitSequence(const string &str,Sequence* s=0,bool reverse=false); //from either GATC or 0101 (as mask: ie double each bit)
  BitSequence(SequenceFileReader &reader,Sequence* s); //from either GATC or 0101 (as mask: ie double each bit)
  BitSequence(const BitSequence& a); //copy
  BitSequence(const word length,Sequence* s,bool _reverse=false);
  BitSequence();//uninited
  ~BitSequence();
  BitSequence* reverseComplement(); //generate a reverse complement version

  friend bool operator==(const BitSequence& a,const BitSequence& b);
  friend class Window;
  friend class HashMethod;
  friend void initHashParams(int bits,BitSequence *pat);
  friend ostream& operator<<(ostream &os,const BitSequence &a);

  vector<pair<SeqPos,SeqPos> > matchRegions;

  //mpi specific
  void mpi_distribute(); //distribute worldwide

  //shared memory specific
  SharedMemType shmUsed;
  string shmAnchor;
  void shm_distribute(); //distribute to other local nodes
  void mmap_msync();

 protected:
  //init
  void initPattern(const string &str);
  void init(const BitSequence& a);

  void compileHashFunc();
  int entropyCheck(vector<int> &entropy);
  word* allocWords();

  SeqPos _length;
  word bit_count,word_count;
  int _hashLength;
  bool isPattern;
  Sequence* seq;
  word* _words;
  bool reverse;
  word counted;

  //for fast hash computation:
  HashMethod* hasher;
};

class HashMethod {
 public:
  typedef vector<pair<int,int> > InputList;
  InputList inputs;
  HashMethod(HashMethod &a);
  HashMethod(BitSequence *s);

  static map<int,int> popmap;

  BitSequence *seq;
  int score,id;

  //scoring state
  vector<int> entropy,used;
  int sources,empties;
  word active;
  friend ostream& operator<<(ostream &os,const HashMethod &a);
  int align(int w); //propose an offset for a new input
  void finalize();
  int maxSources();
  void mutate();
  word hash(Window &w);
  int entropyCheck();
  void removeDuplicates();
  void pruneUseless();
 private:
  static int nextId;
};

//tool for skimming bitsequences through a pattern. binds to BitSequence*
class Window { 
 public:
  BitSequence *seq,*pat;
  SeqPos pos,frame;
  void slide(); //advance 1 base (ie: 2 bits)
  void slide(SeqPos dist); //advance dist bits. (note: bits!!! not bases!!!)
  void slideWord(); //advance 1 whole word

  bool equals(const Window &a) const;
  word eqbases(const Window &a) const;
  word hash();
  Window(BitSequence *s,SeqPos pos,BitSequence *p);
  Window(Location &v,BitSequence *p);
  Window(const UsedInt &i,const Sequence *s,BitSequence *p);
  friend ostream& operator<<(ostream &os, const Window &a);
  friend bool operator<(const Window &a,const Window &b);
  string prettyString();
  string toRawString();

 protected:
  void initslide(SeqPos dist);
  SeqPos frames;
  vector<word> buffer;
 public:
  inline word firstWord(){return buffer[0];}
};

class Sequence {
 public:
  Sequence(); //empty sequence (ie: useless)
  Sequence(string); //from a file
  Sequence(const Sequence &obj); // duplicate
  ~Sequence(); //have to cleanup fwd and rev

  string filename,name;
  SeqPos length() const;
  inline int getId(){return seqID;}
  inline SeqPos extCoords(SeqPos mcoords) const
    {return mcoords<0 ? -1-(rev->length()+mcoords):mcoords;}

  friend bool operator ==(const Sequence&,const Sequence&);
  BitSequence *fwd,*rev;

 protected:
  int seqID;

#ifdef USE_SHM_SYSV
  //sysvipc specific
  key_t sysv_key;
  int sysv_shmid;

  friend class BitSequence; //i need you to be my friend!!
#endif
};

class Location {
 public:
  SeqIdx seqno;
  SeqPos pos;

  Sequence* seq() const;
  inline int seqId() const {return seqno;} //umm this should be "return seqno" i think?
  BitSequence* bitSeq();
  bool equal(const Location &a,BitSequence &patt) const;
  friend bool operator==(const Location &a,const Location &b);
  friend bool operator<(const Location &a,const Location &b);
  friend ostream& operator<<(ostream &os,const Location &a);
  inline SeqPos bitSeqPos(){return (pos>0 ? pos:0-pos)-1;}
  pair<SeqPos,SeqPos>* localRegion();
  vector<pair<SeqPos,SeqPos> >::iterator localRegionIte(); //find region local to pos
  
  Location(Sequence*,SeqPos);
  Location(SeqIdx, SeqPos);
  Location(); //this is dangerous. be careful with it.
};

typedef Location HashVal;
typedef list<Location> LocSubList;
typedef vector<LocSubList> LocList;
bool operator==(const HashVal &a,const HashVal &b);

class Hash {
 public:
  //abstract parts
  virtual void clear() = 0;
  virtual void add(const HashKey key,const HashVal &val) = 0;
  virtual void lookup(HashKey key,const HashVal &a,LocList &locList) = 0;
  virtual void lookup(HashKey key,LocList &locList) = 0;
  virtual bool emptyAt(const HashKey key) = 0;
  virtual word sizeAt(const HashKey key) = 0;
 
  //for computing memory costs
  static const word linear_cost(word);
  static const word bucket_prep_cost(word);

  // dump / load may be overloaded
  virtual void dump(ostream &os);
  virtual void load(istream &is);

  //non virtual stuff that will never change
  void writeHistogram(ostream &of);
  void writeDetailedHistogram(ostream &of);
  BitSequence* hashpat;
  bool sanityCheck();
  bool sanityCheck(word base);
  word sizeAt(const HashKey key,const HashVal &a); //slow but compatible
  
  //the boring junk
  Hash(BitSequence *pat);
  virtual ~Hash() = 0; 
 protected:
  //nada
};

class AnchorSet;

struct ltRegion : binary_function<pair<SeqPos,SeqPos>,pair<SeqPos,SeqPos>,bool> {
  inline bool operator()(const pair<SeqPos,SeqPos> &a, const pair<SeqPos,SeqPos> &b) const {
    return a.second<b.second;
  }
};

class IntervalSet { //intermediate type used for building/modifying AnchorSets
 public:
  IntervalSet();
  IntervalSet(const AnchorSet& a);
  void invert(); //flip everything from forwards<->backwards
  void coalesce(const AnchorSet& neighbor);
  bool contains(const IntervalSet &a);
  bool overlaps(const IntervalSet &a);
  bool overlaps(const AnchorSet &a);
  void add(HashVal&,SeqPos length);
  void extend();
  pair<double,SeqPos> entropy() const;
  
  bool sanityCheck() const;

  vector<UsedInt> spaces;
  HashCount members;
};

class AnchorSet { //anchorsets are necessarily tied to a set of usedmap entries
 public:
  HashCount members;
  vector<MapIte> spaces;

  AnchorSet(const IntervalSet &a,UsedMap& used);
  bool contains(const IntervalSet &a);

  size_t hitCount();
  bool sanityCheck() const;
  double uniqueness() const;
  pair<SeqPos,SeqPos> bitscore() const;

  string asString() const;
  friend ostream& operator<<(ostream &os, const AnchorSet &a);
};

class UsedMap {
 public:
  UsedMap();

  void insert(IntervalSet& a);
  bool sanityCheck();
  string asString();
  int count() const;

  ostream& saveDetails(ostream &os,ostream &bitos); //make second cerr to skip
  void writeOut(ostream &os);

  friend AnchorSet::AnchorSet(const IntervalSet &a,UsedMap& used);
 protected:
  bool alreadyExists(const IntervalSet &a);
  bool merge(IntervalSet& a);
  void add(const IntervalSet& a);
  void remove(AnchorSet *a); //deletes a!

  vector<usedItree > used;
  bool fetchSetIds(int seq,const UsedInt &a,set<AnchorSet*> &out);
  bool fetchNearbySetIds(int seq,const UsedInt &a,set<AnchorSet*> &out);
};

class RepeatMap {
 public:
  void add(const LocList &l);
  void writeOut(ostream &os);

  RepeatMap(int size);
  RepeatMap();
 protected:
  list<LocList> clusters;
};

#endif

